#!/usr/bin/env python
# coding: utf-8

'''
A module for processing two (semi)standardized text variables: snippet.tags, topic Details.topic Categories. They are variables from the output of the YouTube API method `videos`. As well the module creates a column with a video's absolute link
'''

# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        import pandas, warnings
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '').replace('_', '')
        if '.' in module: module = module.split('.')[1] 
        print('Пакет', module,
              'НЕ прединсталлируется с установкой Анаконды, но для работы скрипта требуется этот пакет, поэтому он будет инсталлирован сейчас\n')
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print('Пакет', module
                  , 'НЕ прединсталлируется с установкой Анаконды, для работы скрипта требуется этот пакет,'
                  , 'но инсталлировать его не удаётся, попробуйте инсталлировать его вручную, после чего снова запустите требуемый скрипт пакета\n')
            break

def tag_topic_URL(videoS):
    """
    Функция для обработки двух (полу)стандартизированных текстовых переменных выгрузки метода videos: snippet.tags, topicDetails.topicCategories, а также создания столбца с абсолютной ссылкой
    
    Parameters
    ----------
    videoS : DataFrame
    """    
    warnings.filterwarnings("ignore")
    varS = ['snippet.tags', 'topicDetails.topicCategories']
    for var in varS:
        nanVideoS = videoS[videoS[var].isna()]
        notNanVideoS = videoS[videoS[var].notna()]
        notNanVideoS[var] = notNanVideoS[var].apply(lambda content: ' '.join(content))
        if var == 'snippet.tags': # расщепить на отдельные слова и удалить средли них дубликаты
             notNanVideoS[var] =  notNanVideoS[var].str.split().apply(lambda content: list(dict.fromkeys(content))).apply(lambda content: ' '.join(content))
        if var == 'topicDetails.topicCategories':
            notNanVideoS['https://en.wikipedia.org/wiki/'] = notNanVideoS[var].str.replace('https://en.wikipedia.org/wiki/', '')
        videoS = pandas.concat([nanVideoS, notNanVideoS])
    
    videoS['URL'] = 'https://www.youtube.com/video/' + videoS['id'] # столбец с абсолютной ссылкой
    
    display(videoS[['snippet.tags', 'https://en.wikipedia.org/wiki/', 'URL']].tail())
    print('Число строк', videoS.shape[0])
    return videoS



# coding: utf-8

'''
A module for formatting to a dataframe files of the formats: CSV, Excel and JSON. It facilitates working with data from social media
'''

# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        # from google_drive_downloader import GoogleDriveDownloader
        import os, pandas
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '').replace('_', '')
        if '.' in module: module = module.split('.')[1] 
        # if 'google_drive_downloader' in module: module = 'googledrivedownloader'
        print('Пакет', module
              , 'НЕ прединсталлируется с установкой Анаконды, но для работы скрипта требуется этот пакет, поэтому он будет инсталлирован сейчас\n'
              , 'Попытка №', attempt, 'из 10')
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print('Пакет', module
                  , 'НЕ прединсталлируется с установкой Анаконды, для работы скрипта требуется этот пакет,'
                  , 'но инсталлировать его не удаётся, попробуйте инсталлировать его вручную, после чего снова запустите требуемый скрипт пакета\n')
            break

def getFolder():
    folder = input('--- С какой директорией хотите работать? Укажите полный путь к ней, включая её саму:')
    folder = folder.replace('"', '') # поскольку в Windows путь из Проводника закавычен
    slash = '\\' if os.name == 'nt' else '/' # выбор слэша в зависимости от ОС
    return folder, slash

def getFolderFile(*arg): # арки: fileName и folder
    # print(*arg)
    slash = '\\' if os.name == 'nt' else '/' # выбор слэша в зависимости от ОС
    if len(arg) == 0:
        # print(0)
        folderFile = input('--- С каким файлом хотите работать? Укажите полный путь, включая название файла:')
        folderFile = folderFile.replace('"', '') # поскольку в Windows путь из Проводника закавычен
        # print(folderFile)
        fileName = folderFile.split(slash)[-1]
        # print(fileName)
        folder = slash.join(folderFile.split(slash)[:-1]) # из полного пути убрать имя файла
        folder += slash
        # print(folder)

    if len(arg) == 1:
        # print(1)
        fileName = arg[0]
        fileName = fileName.replace('"', '') # поскольку в Windows путь из Проводника закавычен
        # print(fileName)
        folder = ''
        # print(folder)

    if len(arg) == 2:
        # print(2)
        fileName = arg[0]
        fileName = fileName.replace('"', '') # поскольку в Windows путь из Проводника закавычен
        # print(fileName)
        folder = arg[1]
        folder = folder.replace('"', '') # поскольку в Windows путь из Проводника закавычен
        folder += slash
        # print(folder)
    return fileName, folder, slash

def excel2df(*arg):
    # print(*arg)
    fileName, folder, slash = getFolderFile(*arg)
    folderFile = folder + fileName
    # print(folderFile)
    error = None
    try:
        df = pandas.read_excel(folderFile, index_col=0)
        print(f'Файл "{folderFile}" импортирован')
    except FileNotFoundError:
        errorDescription = sys.exc_info()
        error = str(errorDescription[1]).replace("No module named '", '').replace("'", '').replace('_', '')
        print('error:', error)     
    
    if error == None:
        display(df.head())
        print('Число столбцов:', df.shape[1], ', число строк', df.shape[0], '\n')
    else:
        df = None
    return df, error, fileName, folder, slash    

def files2df(*arg):
    """
    Функция для оформления в датафрейм таблицы из файла формата Excel и присоединения к ней связанных ключом (id) таблиц из файлов форматов CSV, Excel и JSON, расположенных в той же директории
    
    Parameters
    ----------

    """    
    # print(*arg)
    print('Эта функция предназначена для оформления в датафрейм таблицы из файла формата Excel'
          , 'и присоединения к ней связанных ключом (id) таблиц из файлов форматов CSV, Excel и JSON,'
          , 'расположенных в той же директории')
    df, error, fileName, folder, slash = excel2df(*arg)
    
    # print('fileName', fileName) # для отладки
    # print('folder', folder) # для отладки
    if len(folder) == 0: # значит, из excel2df полный путь передан в fileName
        folder = slash.join(fileName.split(slash)[:-1]) + slash # из полного пути убрать имя файла
        fileName = fileName.split(slash)[-1] # из полного пути оставить только имя файла

    fileNameS = os.listdir(folder)
    print(fileNameS)
    fileNameS.remove(fileName)      
    
    formatS = ['XLSX', 'CSV', 'JSON']
    for frmt in formatS:
        print('\n--- Если требуется найти ещё файл формата', frmt, 'для присоединения, то нажмите Enter'
              , '\n--- Если НЕ требуется, то введите любой символ и нажмите Enter')    
        if len(input()) == 0:
            fileNamesForImport = []
            for fileName in fileNameS:
                if ('.' + frmt.lower()) in fileName:
                    print('--- Найден файл', fileName, '. Если он подходит, то нажмите Enter'
                          , '\n--- Если искать дальше, то введите любой символ и нажмите Enter')
                    if len(input()) == 0:
                        fileNamesForImport.append(fileName)
                        print('--- Файл', fileName, 'учтён; если требуется найти ещё файл формата', frmt, ', то нажмите Enter'
                              , '\n--- Если НЕ требуется, то введите любой символ и нажмите Enter')
                        if len(input()) > 0:
                            break

            if len(fileNamesForImport) > 0:
                # print('fileNamesForImport', fileNamesForImport) # для отладки
                # print('folder', folder) # для отладки
                # print('fileName', fileName) # для отладки
                error = None
                try:
                    for fileName in fileNamesForImport:
                        if 'xlsx' in fileName: df = df.merge(pandas.read_excel(f'{folder}{fileName}', index_col=0), on='id', how='outer')
                        if 'csv' in fileName: df = df.merge(pandas.read_csv(f'{folder}{fileName}'), on='id', how='outer')
                        if 'json' in fileName: df = df.merge(pandas.read_json(f'{folder}{fileName}'), on='id', how='outer')
                        # if 'json' in fileName: df = pandas.concat([df, pandas.read_json(f'{folder}{fileName}')], axis=1)
                except FileNotFoundError:
                    errorDescription = sys.exc_info()
                    error = str(errorDescription[1])
                    print('error:', error) 
            else:
                print('Файлы форматов', formatS, 'не найдены в директории\n')
    if error == None:
        display(df.head())
        print('Число столбцов:', df.shape[1], ', число строк', df.shape[0], '\n')
    else:
        df = None
    return df, error, folder

# def googleDriver2local(file_id, dest_path):
#     GoogleDriveDownloader.download_file_from_google_drive(file_id=file_id, dest_path=dest_path)
#     error = None
#     try:
#         df = pandas.read_excel(dest_path, index_col=0)
#     except FileNotFoundError:
#         errorDescription = sys.exc_info()
#         error = str(errorDescription[1]).replace("No module named '", '').replace("'", '').replace('_', '')
#         print('error:', error) 
        
#     if error == None:
#         display(df.head())
#         print('Число столбцов:', df.shape[1], ', число строк', df.shape[0], '\n')
#     else:
#         df = None
#     return df, error



from .utils import *
from .bivariate_association import Crosstab
from .comparison_of_central_tendency import ANOVA

import pandas as pd
import numpy as np
import os
import pydot
from IPython.display import Image, display

from sklearn.metrics import r2_score as r2

class CHAIDClassifier:
    """
    Class to build the CHAID tree on a categorical dependent variable.

    Parameters
    ----------
    method : 'pearson' or 'likelihood' 
        Which Chi-square metric to use
        (CURRENTLY ONLY 'PEARSON' IS AVAILABLE)
    max_depth : int 
        Maximum possible depth of tree
    min_child_node : int 
        Minimum possible number of observations in a child node
    min_parent_node : int 
        Minimum possible number of observations in a parent node
    sig_level_split : float 
        Significance level to split a node
    sig_level_merge : float 
        Significance level to merge categories
    bonferroni : bool 
        (CURRENTLY UNAVAILIABLE) Whether to use a Bonferroni's adjustment for p-values
    allow_resplit : bool 
        (CURRENTLY UNAVAILIABLE) Whether to re-split merged categories
    n_intervals : int 
        Maximum possible number of intervals to bin scale variables

    Attributes
    ----------
    depth : int
        The depth of the obtained tree
    nodes : pd.DataFrame
        A table aggregating information about all nodes of a tree
    terminal_nodes : pd.DataFrame
        A table aggregating information about terminal nodes of a tree
    significant_variables : list
        Names of variables remained in a tree
    classification_table : pd.DataFrame
        A classification table for train data
    precision_and_recall : pd.DataFrame
        A table with precision, recall, and f1-score for train data
    classification_table_test : pd.DataFrame
        A classification table for test data
    precision_and_recall_test : pd.DataFrame
        A table with precision, recall, and f1-score for test data
    """
    def __init__(self,
                method='pearson',
                max_depth=3,
                min_child_node=50,
                min_parent_node=100,
                sig_level_split=0.05,
                sig_level_merge=0.05,
                bonferroni=True,
                allow_resplit=False,
                n_intervals=10):
        
        self.method = method
        self.max_depth = max_depth
        self.min_child_node = min_child_node
        self.min_parent_node = min_parent_node
        self.sig_level_split = sig_level_split
        self.sig_level_merge = sig_level_merge
        self.bonferroni = bonferroni
        self.allow_resplit = allow_resplit
        self.n_intervals = n_intervals
        
        self.nodes = []
        
    def fit(self,
            data,
            dependent_variable,
            independent_variables,
            scale_variables=[],
            ordinal_variables=[],
            test_data=None,
            show_results=True,
            plot_tree=True,
            save_plot=False,
            save_plot_path='',
            save_plot_name='tree',
            save_plot_format='png',
            tree_in_table_format=False,
            n_decimals=3):

        """
        Fit a model to the given data.

        Parameters
        ----------
        data : pd.DataFrame 
            Data to fit a model (train data)
        dependent_variable : str 
            Name of a categorical dependent variable
        independent_variables : list 
            List of names of independent variables
        scale_variables : list 
            Names of independent variables that should
            be considered as scale variables
        ordinal_variables : list 
            Names of independent variables that should
            be considered as ordinal variables; to use this option properly 
            you should either add digits to the labels 
            (i.e., 'low', 'medium', 'high' -> '1. low', '2. medium', '3. high'),
            or turn data into a numeric dtype (1, 2, 3),
            or convert your variable to pandas.Categorical and set the order,
            see more: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Categorical.html
        test_data : pd.DataFrame
            Data to test a model
        show_results : bool 
            Whether to show results of analysis
        plot_tree : bool 
            Whether to plot the obtained tree
        save_plot : bool 
            Whether to save a plot
        save_plot_path : str 
            Folder to save a plot
        save_plot_name :str 
            What name to use when save a plot
        save_plot_format : str 
            What format to use when save a plot ('png', 'svg', 'jpg')
        tree_in_table_format : bool 
            Whether to show tree as a table when showing results
        n_decimals : int 
            Number of digits to round results when showing them

        Returns
        -------
        self
            The current instance of the CHAIDClassifier class
        """
        #todo - handle missing data
        #todo - allow resplit
        #todo - bonferroni
        #todo - different methods in crosstab
        
        self._initial_data = data
        data = data.dropna(subset=independent_variables+[dependent_variable])
        self._data = data.copy()
        self._dependent_variable = dependent_variable
        self._independent_variables = independent_variables
        self._ordinal_variables = ordinal_variables
        self._scale_variables = scale_variables
        self._use_test_data = True if test_data is not None else False
        if self._use_test_data:
            self._test_data = test_data.dropna(subset=independent_variables+[dependent_variable])
        
        self._mode = self._data[dependent_variable].mode().item()

        if self.max_depth is None:
            self.max_depth = len(independent_variables)
        
        tree_nodes = []
        
        tree_nodes.extend(self._build_nodes_from_variable(data,
                                                          dependent_variable))
        
        
        
        tree_nodes = self._split_nodes(data,
                                       dependent_variable,
                                       independent_variables,
                                       tree_nodes)
                                     
        self.nodes = pd.DataFrame(tree_nodes)
        
        
        if len(self.nodes) > 1:
            terminal_nodes_idx = self.nodes.apply(lambda x: self._check_if_terminal_node(x), axis=1)
            self.terminal_nodes = self.nodes[terminal_nodes_idx]
            terminal_nodes_lst = [tree_nodes[i] for i in self.terminal_nodes.index]
            depths = [CHAIDClassifier._get_node_depth(node, tree_nodes)\
                                for node in terminal_nodes_lst]
            self.depth = max(depths)
            
            self._node_interactions = self.get_interactions() 
            self.significant_variables = self.get_significant_variables()
        else:
            self.terminal_nodes = []
            self.depth = 0
            self.significant_variables = []
            self._data[f'{dependent_variable} (predicted)'] = self._data[dependent_variable].mode().item()
        
        self.classification_table = self.get_classification_table()
        self.precision_and_recall = self.get_precision_and_recall()
        
        if self._use_test_data:
            test_predictions = self.predict(self._test_data)
            self.classification_table_test = classification_table(self._test_data[dependent_variable],
            test_predictions[f'{dependent_variable} (predicted)'])
            self.precision_and_recall_test = precision_and_recall(self.classification_table_test)

        if show_results:
            self.show_results(tree_in_table_format=tree_in_table_format, 
                              n_decimals=n_decimals)
        if plot_tree:
            print('------------------\n')
            print('Tree plot')
            self.plot_tree(
                save_plot=save_plot,
                save_plot_path=save_plot_path,
                save_plot_name=save_plot_name,
                save_plot_format=save_plot_format)
        
        return self
    
    def get_significant_variables(self):
        """
        Identify which variables remained in the tree,
        i.e. should be considered as significant ones.

        Returns
        -------
        list
            Names of significant variables
        """

        significant_variables = get_categories(self.nodes['Variable'])
        significant_variables.remove(self._dependent_variable)  
        return significant_variables
    
    def plot_tree(self,
            save_plot=False,
            save_plot_path=None,
            save_plot_name='tree',
            save_plot_format='png'):
        """
        Plot the obtained tree.

        Parameters
        ----------
        save_plot : bool 
            Whether to save a plot
        save_plot_path : str 
            Folder to save a plot
        save_plot_name : str 
            What name to use when save a plot
        save_plot_format : str 
            What format to use when save a plot ('png', 'svg', 'jpg')
        tree_in_table_format : bool 
            Whether to show tree as a table when showing results
        """
        
        graph = pydot.Dot(graph_type='graph', rankdir='LR')

        node_labels = {}

        dep_var = self._dependent_variable
        categories = list(self.nodes.loc[0, 'Dependent variable'].keys())
        colors = dict(zip(categories, available_colors[:len(categories)]))

        #chi2s = []

        for idx, node in self.nodes.iterrows():

            current_node = node['Node']
            node_variable = node['Variable']
            if 'Category' in node:
                node_category = node['Category']
            predicted_category = node['Mode']
            n = node['N']
            if 'Parent node' in node:
                parent_node = node['Parent node']
            else:
                parent_node = None
            if 'Chi-square' in node:
                chi2 = round(node['Chi-square'], 1)
            fillcolor=colors[predicted_category]
            if current_node == 'Node 0':
                node_label = {
                    current_node: f'{current_node}\n{node_variable}\nN = {n}\nMode = {predicted_category}'
                }

            else:
                node_label = {
                    current_node: f'{current_node}\n{node_variable} =\n{node_category}\nN = {n}\n{dep_var} = {predicted_category}'
                }

            node_labels.update(node_label)    

            node_dot = pydot.Node(node_labels[current_node],
                                  style='"rounded,filled"',
                                  fillcolor=fillcolor,
                                  shape='box')
            graph.add_node(node_dot)

            if pd.notnull(parent_node):
                graph.add_edge(pydot.Edge(node_labels[parent_node], node_labels[current_node]))
        if save_plot:
            if save_plot_path is not None:
                path = os.path.join(save_plot_path, f'{save_plot_name}.{save_plot_format}')
            else: 
                path = f'{save_plot_name}.{save_plot_format}'
                
            if save_plot_format.lower().strip()=='png':
                graph.write_png(path, encoding='utf-8')
            elif save_plot_format.lower().strip() in ['jpg', 'jpeg']:
                graph.write_jpg(path, encoding='utf-8')
            elif save_plot_format.lower().strip()=='svg':
                graph.write_svg(path, encoding='utf-8')
            else:
                raise ValueError('Unknown image format. Available options: png, jpg, svg')
                
        display(Image(graph.create_png(encoding='utf-8')))
    
    def get_precision_and_recall(self):
        """
        Estimate precision, recall, and F-score for all the categories.

        Returns
        -------
        pd.DataFrame
            A table with estimated metrics
        """

        results = precision_and_recall(self.classification_table)
        return results
    
    def summary(self):
        """
        Get model summary.

        Returns
        -------
        pd.DataFrame
            A summary table
        """
        statistics = ['Growing Method',
                     'Dependent variable',
                     'Independent variables',
                     'Maximum tree depth',
                     'Minimum cases in child node',
                     'Minimum cases in parent node',
                     'Significant variables',
                     'Number of nodes',
                     'Number of terminal nodes',
                     'Depth']
        
        data = ['CHAID',
                f'{self._dependent_variable}',
                f'{", ".join(self._independent_variables)}',
                f'{self.max_depth}',
                f'{self.min_child_node}',
                f'{self.min_parent_node}',
                f'{", ".join(self.significant_variables)}',
                f'{len(self.nodes)-1}',
                f'{len(self.terminal_nodes)}',
                f'{self.depth}']
        
        results = pd.DataFrame(data, index=statistics, columns=[''])
        
        return results
    
    def show_results(self,
                     tree_in_table_format=False,
                     n_decimals=3):
        """
        Show results of the analysis in a readable form.
        
        Parameters
        ----------
        tree_in_table_format : bool 
            Whether to show tree as a table when showing results
        n_decimals : int 
            Number of digits to round results when showing them
        """
        print('\nCHAID SUMMARY')
        print('------------------\n')
        print('Tree information')
        display(self.summary())
        if tree_in_table_format:
            print('------------------\n')
            print('Tree nodes')
            display(self.nodes.style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption(f'Dependent variable: {self._dependent_variable}, independent variables: {", ".join(self._independent_variables)}'))
        print('------------------\n')
        print('Classification table')
        display(self.classification_table.style\
                .format(precision=1))
        if self._use_test_data:
            print('Classification table for test data')
            display(self.classification_table_test.style\
                    .format(precision=1))
        print('------------------\n')
        print('Prediction quality metrics')
        display(self.precision_and_recall.style\
                .format(precision=n_decimals))
        if self._use_test_data:
            print('Prediction quality metrics for test data')
            display(self.precision_and_recall_test.style\
                .format(precision=n_decimals))
                             
    
    def get_classification_table(self):
        """
        Get the classification table as it is shown in SPSS.

        Returns
        -------
        pd.DataFrame
            A classification table
        """
        dependent_variable = self._dependent_variable
        classification = classification_table(self._data[dependent_variable],
        self._data[f'{dependent_variable} (predicted)']
        )
        return classification
        
    def _check_if_terminal_node(self, node):
        if node['Node'] in self.nodes['Parent node'].tolist() or pd.isna(node['Parent node']):
            return False
        else:
            return True
    
    @staticmethod
    def _predict_one_observation(observation,
                                 nodes,
                                 scale_variables,
                                 constant):
        if observation.hasnans:
            return None, None

        predicted_node = 'Node 0'
        while True:
            current_nodes_level = nodes[nodes['Parent node']==predicted_node]
            if len(current_nodes_level) == 0:
                return predicted_category, predicted_node
            split_variable = current_nodes_level['Variable'].iloc[0]
            observation_category = observation[split_variable]
            if split_variable not in scale_variables:
                try:
                    rule = current_nodes_level.loc[:, 'Category'].str.split(' / ').apply(lambda x: True if str(observation_category) in x else False)
                except AttributeError:
                    rule = current_nodes_level.loc[:, 'Category'].astype(float).astype(str).str.split(' / ').apply(lambda x: True if str(float(observation_category)) in x else False)
            else:
                rule = current_nodes_level.loc[:, 'Category'].apply(lambda x: x.overlaps(pd.Interval(observation_category, observation_category, 'both')))
            
            try:
                predicted_category = current_nodes_level[rule]['Mode'].iloc[0]
                predicted_node = current_nodes_level[rule]['Node'].iloc[0]
            except IndexError:
                predicted_category = constant
                predicted_node = 'Out of tree'
    
    def predict(self,
                data=None,
                dependent_variable=True,
                node=False,
                interaction=False,
                add_to_data=False):
        """
        Predict a value of the dependent variable and/or a node for the given data.

        Parameters
        ----------
        data : pd.DataFrame 
            Data for which values or nodes should be predicted
        dependent_variable : bool 
            Whether to predict the value of the dependent variable
        node : bool 
            Whether to predict the node
        interaction : bool 
            Whether to predict interaction corresponded to the node
        add_to_data : bool 
            Whether to merge predictions with the given data

        Returns
        -------
        pd.DataFrame
            Requested values
        """

        if data is None or data[self._independent_variables]\
        .equals(self._initial_data[self._independent_variables]):
            result = self._data[[f'{self._dependent_variable} (predicted)', 'Node']].copy()
        
        else:
        
            data_for_prediction = data[self.significant_variables].copy()
            result = data_for_prediction.apply(lambda x: CHAIDClassifier._predict_one_observation(
                x, 
                self.nodes, 
                self._scale_variables,
                self._mode
            ), axis=1, result_type='expand')
            result.columns = [f'{self._dependent_variable} (predicted)', 'Node']
                                 
        result['Interaction'] = result['Node'].map(self._node_interactions)

        columns_to_show = []                     
        if dependent_variable:
            columns_to_show.append(f'{self._dependent_variable} (predicted)')
        if node:
            columns_to_show.append('Node')
        if interaction:
            columns_to_show.append('Interaction')

        result = result[columns_to_show].copy()
            
        if add_to_data:
            return pd.concat([data, result], axis=1)
        else:
            return result       
    
    
    @staticmethod
    def _get_most_significant_variable(data, 
                                       dependent_variable,
                                       independent_variables,
                                       method):
        #todo - different methods in crosstab (pearson+likelihood)
        results = []
        for variable in independent_variables:
            if len(data[variable].value_counts()) > 1:
                ctab = Crosstab(data, variable, dependent_variable, show_results=False, only_stats=True)
                results.append([ctab.pvalue, ctab.chi_square, ctab.dof])
            else:
                results.append([1, 0, 0])

        if len(results) == 0:
            return None, None, None, None
        else:
            results = pd.DataFrame(results,
                                   columns=['pvalue', 'chi2', 'dof'],
                                   index=independent_variables)
            most_significant_variable = results['pvalue'].idxmin()
            its_pvalue = results.loc[most_significant_variable, 'pvalue']
            its_chi2 = results.loc[most_significant_variable, 'chi2']
            its_dof = results.loc[most_significant_variable, 'dof']
            return most_significant_variable, its_pvalue, its_chi2, its_dof

    def _merging(self,
                 data,
                 dependent_variable,
                 independent_variable):

        #todo - different methods in crosstab (pearson+likelihood)
        
        sig_level = self.sig_level_merge
        method = self.method
        min_child_node = self.min_child_node
        min_parent_node = self.min_parent_node
        
        # ! todo - resplit categories !
        resplit = self.allow_resplit
        
        data = data.copy()
        
        max_pvalue = 1
        
        while max_pvalue > sig_level:           
            categories = get_categories(data[independent_variable])
            
            if len(categories) <= 2:
                if len(categories) == 2 and min_child_node is not None:
                    min_count = data[independent_variable].value_counts().min()
                    if min_count < min_child_node:
                        if independent_variable not in self._scale_variables:
                            data[independent_variable] = merge_two_cats(data[independent_variable],
                                                                                               categories[0],
                                                                                               categories[1])
                        else:
                            data[independent_variable] = merge_two_intervals(data[independent_variable],
                                                                                              categories[0],
                                                                                              categories[1])
                
                return data[independent_variable]
                 
            if independent_variable not in self._ordinal_variables + self._scale_variables:
                categories_combo = get_unordered_combinations(categories, 2)
            else:
                categories_combo = get_ordered_combinations(categories, 2)
            
            results = pd.DataFrame(columns=categories,
                                   index=categories)
            for pair in categories_combo:
                pair_data = data[data[independent_variable].isin(pair)]
                if len(pair_data) > 0:
                    ctab = Crosstab(pair_data, independent_variable, dependent_variable, only_stats=True) #add method
                    var1, var2 = pair
                    results.loc[var1, var2] = ctab.pvalue
                    results.loc[var2, var1] = ctab.pvalue
            max_pvalue = results.max().max()
            if max_pvalue > sig_level:
                cat1 = results.max().idxmax()
                cat2 = results[cat1][results[cat1]==max_pvalue].index[0]
                if independent_variable not in self._scale_variables:
                    data[independent_variable] = merge_two_cats(data[independent_variable],
                                                                                       cat1,
                                                                                       cat2)
                else:
                    data[independent_variable] = merge_two_intervals(data[independent_variable],
                                                                                      cat1,
                                                                                      cat2)
        
        if min_child_node is not None:
            min_count = data[independent_variable].value_counts().min()
            n_categories = len(data[independent_variable].unique())
    #         (Optional) Any category having too few observations (as compared with a user-specified minimum segment size) 
    #         is merged with the most similar other category as measured by the largest of the p-values.

            while 0 < min_count < min_child_node and n_categories > 1:                

                min_count_category = data[independent_variable].value_counts().idxmin()
                categories = get_categories(data[independent_variable])
                if independent_variable not in self._ordinal_variables + self._scale_variables:
                    categories_combo = get_unordered_combinations(categories, 2)
                else:
                    categories_combo = get_ordered_combinations(categories, 2)
                results = pd.DataFrame(columns=categories,
                       index=categories)
                for pair in categories_combo:
                    pair_data = data[data[independent_variable].isin(pair)]
                    if len(pair_data) > 0:
                        ctab = Crosstab(pair_data, independent_variable, dependent_variable, only_stats=True) #add method
                        var1, var2 = pair
                        results.loc[var1, var2] = ctab.pvalue
                        results.loc[var2, var1] = ctab.pvalue

                max_pvalue = results[min_count_category].max()
                max_pvalue_category = pd.to_numeric(results[min_count_category]).idxmax()
                if independent_variable not in self._scale_variables:
                    data[independent_variable] = merge_two_cats(data[independent_variable],
                                                                                 min_count_category,
                                                                                 max_pvalue_category)
                else:
                    data[independent_variable] = merge_two_intervals(data[independent_variable],
                                                                                 min_count_category,
                                                                                 max_pvalue_category)          
                min_count = data[independent_variable].value_counts().min()
                n_categories = len(data[independent_variable].unique())
        return data[independent_variable]

    def _build_nodes_from_variable(self,
                                   data,
                                   dependent_variable,
                                   independent_variable=None,
                                   start_counter=0,
                                   parent_node=None):
        
        if independent_variable is not None:
            categories = get_categories(data[independent_variable])
            nodes = []
            counter = start_counter
            n_observations = []
            for category in categories:
                filter_data = data[data[independent_variable]==category][dependent_variable]
                n = len(filter_data)
                n_observations.append(n)
            small_n = [n for n in n_observations if n <= self.min_child_node]   
            if len(small_n) < len(categories) - 1:
                for category in categories:
                    filter_data = data[data[independent_variable]==category][dependent_variable]
                    n = len(filter_data)
                    dependent_variable_dist = dict((round(filter_data.value_counts(normalize=True, sort=False)*100, 3).items()))
                    mode = filter_data.value_counts().idxmax()
                    node_info = {'Node': f'Node {counter}',
                                 'Parent node': parent_node,
                                 'Variable': independent_variable,
                                 'Category': category,
                                 'Dependent variable': dependent_variable_dist,
                                 'Mode': mode,
                                 'N': n}
                    nodes.append(node_info)
                    counter += 1
        else:
            filter_data = data[dependent_variable]
            n = len(filter_data)
            dependent_variable_dist = dict((round(filter_data.value_counts(normalize=True, sort=False)*100, 3).items()))
            mode = filter_data.value_counts().idxmax()
            node_info = {'Node': f'Node 0',
                         'Variable': dependent_variable,
                         'Dependent variable': dependent_variable_dist,
                         'Mode': mode,
                         'N': n}
            nodes = [node_info]                                
                
        return nodes
    
    def _split_nodes(self,
                     data,
                     dependent_variable,
                     independent_variables,
                     tree_nodes,
                     parent_node='Node 0'):
        
        data_iter = data.copy()
        for variable in independent_variables:
            
            if variable in self._scale_variables:
                
                data_iter[variable] = binning(data_iter[variable], self.n_intervals)
            
            data_iter[variable] = self._merging(data_iter,
                                                dependent_variable,
                                                variable)
        
        split_variable, split_pvalue, split_chi2, split_dof = CHAIDClassifier\
        ._get_most_significant_variable(data_iter,
                                        dependent_variable,
                                        independent_variables,
                                        self.method)
        
        if split_variable is not None:
            split_info = {
                         'Chi-square': split_chi2,
                         'p-value': split_pvalue,
                         'dof': split_dof}
            if split_info['p-value'] <= self.sig_level_split:
                iter_nodes = self._build_nodes_from_variable(data_iter,
                                                                       dependent_variable,
                                                                       split_variable,
                                                                       start_counter=len(tree_nodes),
                                                                       parent_node=parent_node)
                for node in iter_nodes:
                    node.update(split_info)
                
                if len(iter_nodes) >= 2:
                    tree_nodes.extend(iter_nodes)
                    current_depths = [CHAIDClassifier._get_node_depth(node, tree_nodes)\
                                        for node in iter_nodes if node['Node']!='Node 0']
                    current_depth = max(current_depths)
                    if current_depth < self.max_depth:    
                        
                        for i in range(len(iter_nodes)):
                            parent_node = iter_nodes[i]['Node']
                            node_variable = iter_nodes[i]['Variable']
                            node_category = iter_nodes[i]['Category']
                            node_prediction = iter_nodes[i]['Mode']
                            data_per_node = data[data_iter[node_variable]==node_category].copy()
                            self._data.loc[data_per_node.index, f'{self._dependent_variable} (predicted)'] = node_prediction
                            self._data.loc[data_per_node.index, 'Node'] = parent_node
                            if len(data_per_node) >= self.min_parent_node:
                                available_variables = independent_variables.copy()
                                available_variables.remove(node_variable)
                                tree_nodes = self._split_nodes(data_per_node,
                                                                   dependent_variable,
                                                                   available_variables,
                                                                   tree_nodes,
                                                                   parent_node)
                    else:
                        for i in range(len(iter_nodes)):
                            parent_node = iter_nodes[i]['Node']
                            node_variable = iter_nodes[i]['Variable']
                            node_category = iter_nodes[i]['Category']
                            node_prediction = iter_nodes[i]['Mode']
                            data_per_node = data[data_iter[node_variable]==node_category].copy()
                            self._data.loc[data_per_node.index, f'{self._dependent_variable} (predicted)'] = node_prediction
                            self._data.loc[data_per_node.index, 'Node'] = parent_node
        

        return tree_nodes
    
    #move this method to the base chaid tree class
    @staticmethod
    def _get_node_depth(node, nodes):
        depth = 0
        while True:
            if 'Parent node' in node:
                parent_node = node['Parent node']
                node = nodes[int(parent_node.split()[-1])]
                depth += 1
            else:
                break
        return depth
            
    @staticmethod
    def _get_one_node_interaction(node, nodes):
        interaction = []
        if node['Node'] == 'Node 0':
            return np.nan
        while True:
            if node['Node'] == 'Node 0':
                break
            interaction.append(f'{node["Variable"]} = {node["Category"]}')
            node = nodes[nodes['Node']==node['Parent node']].iloc[0]
        return ' * '.join(interaction)
                               
    def get_interactions(self, result='dict'):
        """
        Return a dictionary or a DataFrame with nodes and interactions corresponded to them.

        Parameters
        ----------
        result : 'dict' or 'DataFrame'
            Data type of results

        Returns
        -------
        dict or pd.DataFrame
            Definitions of nodes in terms of interactions
        """
                               
        results = self.nodes.apply(lambda x: CHAIDClassifier._get_one_node_interaction(x, self.nodes), axis=1)
        results.index = self.nodes['Node']
        if result.lower() == 'dict':
            return dict(results)
        elif result.lower() == 'dataframe':
            return pd.DataFrame(results, columns=['Interaction'])
        else:
            raise ValueError("Unknown result type. Possible values: 'dict' and 'DataFrame'.")


class CHAIDRegressor:
    """
    Class to build the CHAID tree on a scale dependent variable.

    Parameters
    ----------
    max_depth : int 
        Maximum possible depth of tree
    min_child_node : int 
        Minimum possible number of observations in a child node
    min_parent_node : int 
        Minimum possible number of observations in a parent node
    sig_level_split : float 
        Significance level to split a node
    sig_level_merge : float 
        Significance level to merge categories
    bonferroni : bool 
        (CURRENTLY UNAVAILIABLE) Whether to use a Bonferroni's adjustment for p-values
    allow_resplit : bool 
        (CURRENTLY UNAVAILIABLE) Whether to re-split merged categories
    n_intervals : int 
        Maximum possible number of intervals to bin scale variables

    Attributes
    ----------
    depth : int
        The depth of the obtained tree
    nodes : pd.DataFrame
        A table aggregating information about all nodes of a tree
    terminal_nodes : pd.DataFrame
        A table aggregating information about terminal nodes of a tree
    significant_variables : list
        Names of variables remained in a tree
    r2 : float
        R2 score on train data
    r2_test : float 
        R2 score on test data
    """
    def __init__(self,
                 max_depth=3,
                min_child_node=50,
                min_parent_node=100,
                sig_level_split=0.05,
                sig_level_merge=0.05,
                bonferroni=True,
                allow_resplit=False,
                n_intervals=10):
        
        self.max_depth = max_depth
        self.min_child_node = min_child_node
        self.min_parent_node = min_parent_node
        self.sig_level_split = sig_level_split
        self.sig_level_merge = sig_level_merge
        self.bonferroni = bonferroni
        self.allow_resplit = allow_resplit
        self.n_intervals = n_intervals
        self.nodes = []
        
    def fit(self,
            data,
            dependent_variable,
            independent_variables,
            scale_variables=[],
            ordinal_variables=[],
            test_data=None,
            show_results=True,
            plot_tree=True,
            save_plot=False,
            save_plot_path='',
            save_plot_name='tree',
            save_plot_format='png',
            tree_in_table_format=False,
            n_decimals=3):
        """
        Fit model to the given data.

        Parameters
        ----------
        data : pd.DataFrame 
            Data to fit a model (train data)
        dependent_variable : str 
            Name of a categorical dependent variable
        independent_variables : list 
            List of names of independent variables
        scale_variables : list 
            Names of independent variables that should
            be considered as scale variables
        ordinal_variables : list 
            Names of independent variables that should
            be considered as ordinal variables; to use this option properly 
            you should either add digits to the labels 
            (i.e., 'low', 'medium', 'high' -> '1. low', '2. medium', '3. high'),
            or turn data into a numeric dtype (1, 2, 3),
            or convert your variable to pandas.Categorical and set the order,
            see more: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Categorical.html
        test_data : pd.DataFrame
            Data to test a model
        show_results : bool 
            Whether to show results of analysis
        plot_tree : bool 
            Whether to plot the obtained tree
        save_plot : bool 
            Whether to save a plot
        save_plot_path : str 
            Folder to save a plot
        save_plot_name :str 
            What name to use when save a plot
        save_plot_format : str 
            What format to use when save a plot ('png', 'svg', 'jpg')
        tree_in_table_format : bool 
            Whether to show tree as a table when showing results
        n_decimals : int 
            Number of digits to round results when showing them

        Returns
        -------
        self
            The current instance of the CHAIDClassifier class
        """
        #todo - handle missing data
        #todo - allow resplit
        #todo - bonferroni
        #todo - different methods in crosstab
        
        self._initial_data = data
        data = data.dropna(subset=independent_variables+[dependent_variable])
        self._data = data.copy()
        self._dependent_variable = dependent_variable
        self._independent_variables = independent_variables
        self._ordinal_variables = ordinal_variables
        self._scale_variables = scale_variables
        self._use_test_data = True if test_data is not None else False
        if self._use_test_data:
            self._test_data = test_data.dropna(subset=independent_variables+[dependent_variable])

        self._mean = self._data[dependent_variable].mean()

        if self.max_depth is None:
            self.max_depth = len(independent_variables)
        
        tree_nodes = []
        
        tree_nodes.extend(self._build_nodes_from_variable(data,
                                                          dependent_variable))
        #print(tree_nodes)
        
        
        tree_nodes = self._split_nodes(data,
                                       dependent_variable,
                                       independent_variables,
                                       tree_nodes)
                                     

        self.nodes = pd.DataFrame(tree_nodes)

        if len(self.nodes) > 1:
            terminal_nodes_idx = self.nodes.apply(lambda x: self._check_if_terminal_node(x), axis=1)
            self.terminal_nodes = self.nodes[terminal_nodes_idx]
            terminal_nodes_lst = [tree_nodes[i] for i in self.terminal_nodes.index]
            depths = [CHAIDRegressor._get_node_depth(node, tree_nodes)\
                                for node in terminal_nodes_lst]
            self.depth = max(depths)

            self._node_interactions = self.get_interactions()      
            self.significant_variables = self.get_significant_variables()
        else:
            self.terminal_nodes = []
            self.depth = 0
            self.significant_variables = []
            self._data[f'{dependent_variable} (predicted)'] = self._data[dependent_variable].mean()
        
        self.r2 = r2(self._data[dependent_variable],
                    self._data[f'{dependent_variable} (predicted)'])
        
        if self._use_test_data:
            test_predictions = self.predict(self._test_data)
            self.r2_test = r2(self._test_data[dependent_variable],
                    test_predictions[f'{dependent_variable} (predicted)'])

        if show_results:
            self.show_results(tree_in_table_format=tree_in_table_format, 
                              n_decimals=n_decimals)
        if plot_tree:
            print('------------------\n')
            print('Tree plot')
            self.plot_tree(
                save_plot=save_plot,
                save_plot_path=save_plot_path,
                save_plot_name=save_plot_name,
                save_plot_format=save_plot_format)
        
        return self
    
    def get_significant_variables(self):
        """
        Identify which variables remained in the tree,
        i.e. should be considered as significant ones.

        Returns
        -------
        list
            Names of significant variables
        """
        significant_variables = get_categories(self.nodes['Variable'])
        significant_variables.remove(self._dependent_variable)  
        return significant_variables
        
    def plot_tree(self,
            save_plot=False,
            save_plot_path=None,
            save_plot_name='tree',
            save_plot_format='png'):
        """
        Plot the obtained tree.

        Parameters
        ----------
        save_plot : bool 
            Whether to save a plot
        save_plot_path : str 
            Folder to save a plot
        save_plot_name : str 
            What name to use when save a plot
        save_plot_format : str 
            What format to use when save a plot ('png', 'svg', 'jpg')
        tree_in_table_format : bool 
            Whether to show tree as a table when showing results
        """
        
        graph = pydot.Dot(graph_type='graph', rankdir='LR')

        node_labels = {}

        dep_var = self._dependent_variable

        for idx, node in self.nodes.iterrows():

            current_node = node['Node']
            node_variable = node['Variable']
            node_category = node['Category']
            predicted_category = node['Mean']
            std = node['St. dev.']
            n = node['N']
            parent_node = node['Parent node']
            
            if current_node == 'Node 0':
                node_0_ci = confidence_interval_mean(predicted_category, std, n, sig_level=self.sig_level_split)
                predicted_category = round(predicted_category, 2)
                std = round(std, 2)
                fillcolor='lightgray'
                #fillcolor=colors[predicted_category]
                node_label = {
                    current_node: f'{current_node}\n{node_variable}\nN = {n}\nMean = {predicted_category}\n(std = {std})'
                }

            else:
                node_x_ci = confidence_interval_mean(predicted_category, std, n, sig_level=self.sig_level_split)
                if confidence_interval_comparison(node_x_ci, node_0_ci)=='overlap':                    
                    fillcolor = 'lightgray'
                elif confidence_interval_comparison(node_x_ci, node_0_ci)=='lower':
                    fillcolor = available_colors[-2]
                elif confidence_interval_comparison(node_x_ci, node_0_ci)=='bigger':
                    fillcolor = available_colors[-1]
                    
                predicted_category = round(predicted_category, 2)
                std = round(std, 2)
                node_label = {
                    current_node: f'{current_node}\n{node_variable} =\n{node_category}\nN = {n}\n{dep_var} = {predicted_category}\n(std = {std})'
                }

            node_labels.update(node_label)    

            node_dot = pydot.Node(node_labels[current_node],
                                  style='"rounded,filled"',
                                  fillcolor=fillcolor,
                                  shape='box')
            graph.add_node(node_dot)

            if pd.notnull(parent_node):
                graph.add_edge(pydot.Edge(node_labels[parent_node], node_labels[current_node]))
        if save_plot:
            if save_plot_path is not None:
                path = os.path.join(save_plot_path, f'{save_plot_name}.{save_plot_format}')
            else: 
                path = f'{save_plot_name}.{save_plot_format}'
                
            if save_plot_format.lower().strip()=='png':
                graph.write_png(path, encoding='utf-8')
            elif save_plot_format.lower().strip() in ['jpg', 'jpeg']:
                graph.write_jpg(path, encoding='utf-8')
            elif save_plot_format.lower().strip()=='svg':
                graph.write_svg(path, encoding='utf-8')
            else:
                raise ValueError('Unknown image format. Available options: png, jpg, svg')
                
        display(Image(graph.create_png(encoding='utf-8')))
    
    
    def summary(self):
        """
        Get model summary.

        Returns
        -------
        pd.DataFrame
            A summary table
        """
        statistics = ['Growing Method',
                     'Dependent variable',
                     'Independent variables',
                     'Maximum tree depth',
                     'Minimum cases in child node',
                     'Minimum cases in parent node',
                     'Significant variables',
                     'Number of nodes',
                     'Number of terminal nodes',
                     'Depth',
                     '',
                     'R2']
        
        data = ['CHAID',
                f'{self._dependent_variable}',
                f'{", ".join(self._independent_variables)}',
                f'{self.max_depth}',
                f'{self.min_child_node}',
                f'{self.min_parent_node}',
                f'{", ".join(self.significant_variables)}',
                f'{len(self.nodes)-1}',
                f'{len(self.terminal_nodes)}',
                f'{self.depth}',
               '',
               f'{round(self.r2, 3)}']
        
        results = pd.DataFrame(data, index=statistics, columns=[''])
        
        if self._use_test_data:
            results.loc['R2 (test data)'] = round(self.r2_test, 3)

        return results
    
    def show_results(self,
                     tree_in_table_format=False,
                     n_decimals=3):
        """
        Show results of the analysis in a readable form.
        
        Parameters
        ----------
        tree_in_table_format : bool 
            Whether to show tree as a table when showing results
        n_decimals : int 
            Number of digits to round results when showing them
        """
        
        print('\nCHAID SUMMARY')
        print('------------------\n')
        print('Tree information')
        display(self.summary())
        if tree_in_table_format:
            print('------------------\n')
            print('Tree nodes')
            display(self.nodes.style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption(f'Dependent variable: {self._dependent_variable}, independent variables: {", ".join(self._independent_variables)}'))
                                 
    def _check_if_terminal_node(self, node):
        if node['Node'] in self.nodes['Parent node'].tolist() or pd.isna(node['Parent node']):
            return False
        else:
            return True
    
    @staticmethod
    def _predict_one_observation(observation,
                                 nodes,
                                 scale_variables,
                                 constant):
        if observation.hasnans:
            return None, None

        predicted_node = 'Node 0'
        
        while True:
            current_nodes_level = nodes[nodes['Parent node']==predicted_node]
            #display(current_nodes_level)
            if len(current_nodes_level) == 0:
                return predicted_category, predicted_node
            split_variable = current_nodes_level['Variable'].iloc[0]
            #print(split_variable)
            observation_category = observation[split_variable]
            #print(observation_category)
            if split_variable not in scale_variables:
                try:
                    rule = current_nodes_level.loc[:, 'Category'].str.split(' / ').apply(lambda x: True if str(observation_category) in x else False)
                except AttributeError:
                    rule = current_nodes_level.loc[:, 'Category'].astype(float).astype(str).str.split(' / ').apply(lambda x: True if str(float(observation_category)) in x else False)
            else:
                rule = current_nodes_level.loc[:, 'Category'].apply(lambda x: x.overlaps(pd.Interval(observation_category, observation_category, 'both')))
            try:
                predicted_category = current_nodes_level[rule]['Mean'].iloc[0]
                #print(predicted_category)
                predicted_node = current_nodes_level[rule]['Node'].iloc[0]
            except IndexError:
                predicted_category = constant
                predicted_node = 'Out of tree'
            #print(predicted_node)
    
    def predict(self,
                data=None,
                dependent_variable=True,
                node=False,
                interaction=False,
                add_to_data=False):
        """
        Predict a value of the dependent variable and/or a node for the given data.

        Parameters
        ----------
        data : pd.DataFrame 
            Data for which values or nodes should be predicted
        dependent_variable : bool 
            Whether to predict the value of the dependent variable
        node : bool 
            Whether to predict the node
        interaction : bool 
            Whether to predict interaction corresponded to the node
        add_to_data : bool 
            Whether to merge predictions with the given data

        Returns
        -------
        pd.DataFrame
            Requested values
        """
        
        if data is None:
            result = self._data[[f'{self._dependent_variable} (predicted)', 'Node']].copy()
        
        else:
        
            data_for_prediction = data[self.significant_variables]
            result = data_for_prediction.apply(lambda x: CHAIDRegressor._predict_one_observation(
                x, 
                self.nodes, 
                self._scale_variables,
                self._mean
            ), axis=1, result_type='expand')
            result.columns = [f'{self._dependent_variable} (predicted)', 'Node']
                                 
        result['Interaction'] = result['Node'].map(self._node_interactions)

        columns_to_show = []                     
        if dependent_variable:
            columns_to_show.append(f'{self._dependent_variable} (predicted)')
        if node:
            columns_to_show.append('Node')
        if interaction:
            columns_to_show.append('Interaction')

        result = result[columns_to_show]
            
        if add_to_data:
            return pd.concat([data, result], axis=1)
        else:
            return result       
    
    @staticmethod
    def _get_most_significant_variable(data, 
                                       dependent_variable,
                                       independent_variables):
        results = []
        for variable in independent_variables:
            if len(data[variable].value_counts()) > 1:
                anv = ANOVA(data, dependent_variable, variable, show_results=False)
                results.append([anv.pvalue, anv.F, anv.dof_b, anv.dof_w])
            else:
                results.append([1, 0, 0, 0])

        if len(results) == 0:
            return None, None, None, None, None
        else:
            results = pd.DataFrame(results,
                                   columns=['pvalue', 'F', 'dof_b', 'dof_w'],
                                   index=independent_variables)

            most_significant_variable = results['pvalue'].idxmin()
            its_pvalue = results.loc[most_significant_variable, 'pvalue']
            its_chi2 = results.loc[most_significant_variable, 'F']
            its_dof_b = results.loc[most_significant_variable, 'dof_b']
            its_dof_w = results.loc[most_significant_variable, 'dof_w']

            return most_significant_variable, its_pvalue, its_chi2, its_dof_b, its_dof_w

        
    def _merging(self,
                 data,
                 dependent_variable,
                 independent_variable):
        
        sig_level = self.sig_level_merge
        min_child_node = self.min_child_node
        min_parent_node = self.min_parent_node
        
        # ! todo - resplit categories !
        resplit = self.allow_resplit
        
        data = data.copy()
        
        max_pvalue = 1
        
        while max_pvalue > sig_level:            
            categories = get_categories(data[independent_variable])
            
            if len(categories) <= 2:
                if len(categories) == 2 and min_child_node is not None:
                    min_count = data[independent_variable].value_counts().min()
                    if min_count < min_child_node:
                        if independent_variable not in self._scale_variables:
                            data[independent_variable] = merge_two_cats(data[independent_variable],
                                                                                               categories[0],
                                                                                               categories[1])
                        else:
                            data[independent_variable] = merge_two_intervals(data[independent_variable],
                                                                                              categories[0],
                                                                                              categories[1])
                return data[independent_variable]
                 
            if independent_variable not in self._ordinal_variables + self._scale_variables:
                categories_combo = get_unordered_combinations(categories, 2)
            else:
                categories_combo = get_ordered_combinations(categories, 2)
            
            results = pd.DataFrame(columns=categories,
                                   index=categories)
            for pair in categories_combo:
                pair_data = data[data[independent_variable].isin(pair)]
                if len(pair_data) > 0:
                    anv = ANOVA(pair_data, dependent_variable, independent_variable, show_results=False)
                    var1, var2 = pair
                    results.loc[var1, var2] = anv.pvalue
                    results.loc[var2, var1] = anv.pvalue
            max_pvalue = results.max().max()
            if max_pvalue > sig_level:
                cat1 = results.max().idxmax()
                cat2 = results[cat1][results[cat1]==max_pvalue].index[0]
                if independent_variable not in self._scale_variables:
                    data[independent_variable] = merge_two_cats(data[independent_variable],
                                                                                       cat1,
                                                                                       cat2)
                else:
                    data[independent_variable] = merge_two_intervals(data[independent_variable],
                                                                                      cat1,
                                                                                      cat2)
            
        if min_child_node is not None:
            min_count = data[independent_variable].value_counts().min()
            n_categories = len(data[independent_variable].unique())
    #         (Optional) Any category having too few observations (as compared with a user-specified minimum segment size) 
    #         is merged with the most similar other category as measured by the largest of the p-values.

            while 0 < min_count < min_child_node and n_categories > 1:                

                min_count_category = data[independent_variable].value_counts().idxmin()
                categories = get_categories(data[independent_variable])
                if independent_variable not in self._ordinal_variables + self._scale_variables:
                    categories_combo = get_unordered_combinations(categories, 2)
                else:
                    categories_combo = get_ordered_combinations(categories, 2)
                results = pd.DataFrame(columns=categories,
                       index=categories)
                for pair in categories_combo:
                    pair_data = data[data[independent_variable].isin(pair)]
                    if len(pair_data) > 0:
                        anv = ANOVA(pair_data, dependent_variable, independent_variable, show_results=False)
                        var1, var2 = pair
                        results.loc[var1, var2] = anv.pvalue
                        results.loc[var2, var1] = anv.pvalue

                max_pvalue = results[min_count_category].max()
                max_pvalue_category = pd.to_numeric(results[min_count_category]).idxmax()
                if independent_variable not in self._scale_variables:
                    data[independent_variable] = merge_two_cats(data[independent_variable],
                                                                                 min_count_category,
                                                                                 max_pvalue_category)
                else:
                    data[independent_variable] = merge_two_intervals(data[independent_variable],
                                                                                 min_count_category,
                                                                                 max_pvalue_category)          
                min_count = data[independent_variable].value_counts().min()
                n_categories = len(data[independent_variable].unique())
        return data[independent_variable]

    def _build_nodes_from_variable(self,
                                   data,
                                   dependent_variable,
                                   independent_variable=None,
                                   start_counter=0,
                                   parent_node=None):
        
        if independent_variable is not None:
            categories = get_categories(data[independent_variable])
            nodes = []
            counter = start_counter
            n_observations = []
            for category in categories:
                filter_data = data[data[independent_variable]==category][dependent_variable]
                n = len(filter_data)
                n_observations.append(n)
            small_n = [n for n in n_observations if n <= self.min_child_node]    
            if len(small_n) < len(categories) - 1:
                for category in categories:
                    filter_data = data[data[independent_variable]==category][dependent_variable]
                    n = len(filter_data)
                    mean = filter_data.mean()
                    std = filter_data.std()
                    node_info = {'Node': f'Node {counter}',
                                 'Parent node': parent_node,
                                 'Variable': independent_variable,
                                 'Category': category,
                                 'Mean': mean,
                                 'St. dev.': std,
                                 'N': n}
                    nodes.append(node_info)
                    counter += 1
        else:
            filter_data = data[dependent_variable]
            n = len(filter_data)
            mean = filter_data.mean()
            std = filter_data.std()
            node_info = {'Node': 'Node 0',
                         'Variable': dependent_variable,
                         'Mean': mean,
                         'St. dev.': std,
                         'N': n}
            nodes = [node_info]                                
                
        return nodes
    
    def _split_nodes(self,
                     data,
                     dependent_variable,
                     independent_variables,
                     tree_nodes,
                     parent_node='Node 0'):
        
        data_iter = data.copy()
        for variable in independent_variables:
            
            if variable in self._scale_variables:
                data_iter[variable] = binning(data_iter[variable], self.n_intervals)

            data_iter[variable] = self._merging(data_iter,
                                                dependent_variable,
                                                variable)

        split_variable, split_pvalue, split_F, split_dof_b, split_dof_w = CHAIDRegressor\
        ._get_most_significant_variable(data_iter,
                                        dependent_variable,
                                        independent_variables)
        
        if split_variable is not None:
            split_info = {
                         'F': split_F,
                         'p-value': split_pvalue,
                         'dof_b': split_dof_b,
            'dof_w': split_dof_w}

            if split_info['p-value'] <= self.sig_level_split:
                iter_nodes = self._build_nodes_from_variable(data_iter,
                                                                       dependent_variable,
                                                                       split_variable,
                                                                       start_counter=len(tree_nodes),
                                                                       parent_node=parent_node)
                for node in iter_nodes:
                    node.update(split_info)
                                 
                if len(iter_nodes) >= 2:
                    tree_nodes.extend(iter_nodes)
                    current_depths = [CHAIDRegressor._get_node_depth(node, tree_nodes)\
                                        for node in iter_nodes if node['Node']!='Node 0']
                    current_depth = max(current_depths)
                    if current_depth < self.max_depth:
                        for i in range(len(iter_nodes)):
                            parent_node = iter_nodes[i]['Node']
                            node_variable = iter_nodes[i]['Variable']
                            node_category = iter_nodes[i]['Category']
                            node_prediction = iter_nodes[i]['Mean']
                            data_per_node = data[data_iter[node_variable]==node_category].copy()
                            self._data.loc[data_per_node.index, f'{self._dependent_variable} (predicted)'] = node_prediction
                            self._data.loc[data_per_node.index, 'Node'] = parent_node
                            if len(data_per_node) >= self.min_parent_node:
                                available_variables = independent_variables.copy()
                                available_variables.remove(node_variable)
                                tree_nodes = self._split_nodes(data_per_node,
                                                                   dependent_variable,
                                                                   available_variables,
                                                                   tree_nodes,
                                                                   parent_node)
                    else:
                        for i in range(len(iter_nodes)):
                            parent_node = iter_nodes[i]['Node']
                            node_variable = iter_nodes[i]['Variable']
                            node_category = iter_nodes[i]['Category']
                            node_prediction = iter_nodes[i]['Mean']
                            data_per_node = data[data_iter[node_variable]==node_category].copy()
                            self._data.loc[data_per_node.index, f'{self._dependent_variable} (predicted)'] = node_prediction
                            self._data.loc[data_per_node.index, 'Node'] = parent_node
        

        return tree_nodes
                             
    @staticmethod
    def _get_node_depth(node, nodes):
        depth = 0
        while True:
            if 'Parent node' in node:
                parent_node = node['Parent node']
                node = nodes[int(parent_node.split()[-1])]
                depth += 1
            else:
                break
        return depth
    
    @staticmethod
    def _get_one_node_interaction(node, nodes):
        interaction = []
        if node['Node'] == 'Node 0':
            return np.nan
        while True:
            if node['Node'] == 'Node 0':
                break
            interaction.append(f'{node["Variable"]} = {node["Category"]}')
            node = nodes[nodes['Node']==node['Parent node']].iloc[0]
        return ' * '.join(interaction)
                               
    def get_interactions(self, result='dict'):
        """
        Return a dictionary or a DataFrame with nodes and interactions corresponded to them.

        Parameters
        ----------
        result : 'dict' or 'DataFrame'
            Data type of results

        Returns
        -------
        dict or pd.DataFrame
            Definitions of nodes in terms of interactions
        """
                               
        results = self.nodes.apply(lambda x: CHAIDRegressor._get_one_node_interaction(x, self.nodes), axis=1)
        results.index = self.nodes['Node']
        if result.lower() == 'dict':
            return dict(results)
        elif result.lower() == 'dataframe':
            return pd.DataFrame(results, columns=['Interaction'])
        else:
            raise ValueError("Unknown result type. Possible values: 'dict' and 'DataFrame'.")



#!/usr/bin/env python
# coding: utf-8

'''
A module for an unstandardized text preprocessing
'''

# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        import os, pandas, pymystem3, re, stop_words, time
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '') #.replace('_', '')
        if '.' in module: module = module.split('.')[1] 
        print('Пакет', module,
              'НЕ прединсталлируется с установкой Анаконды, но для работы скрипта требуется этот пакет, поэтому он будет инсталлирован сейчас\n')
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print('Пакет', module
                  , 'НЕ прединсталлируется с установкой Анаконды, для работы скрипта требуется этот пакет,'
                  , 'но инсталлировать его не удаётся, попробуйте инсталлировать его вручную, после чего снова запустите требуемый скрипт пакета\n')
            break

def multispaceCleaner(text):
    cleaned_text = text
    while '  ' in text: cleaned_text = text.replace('  ', ' ')
    while cleaned_text[0] == ' ': cleaned_text = cleaned_text[1:] # избавиться от пробелов в начале текста 
    while cleaned_text[-1] == ' ': cleaned_text = cleaned_text[:-1] # избавиться от пробелов в конце текста 
    return cleaned_text

def pymystemLemmatizer(dfIn, columnWithText):
    """
    Функция для лемматизации текстов пакетом pymystem3
    
    Parameters
    ----------
              dfIn : DataFrame
    columnWithText : str
    """
    df = dfIn.copy()
    time.sleep(0.01)
    print('                                                                                          ', end='\r')
    mstem = pymystem3.Mystem()
    separator = r'|||'
    while len(df[df[columnWithText].str.contains(separator)]) == 0:
        print('--- Не получается использовать разделитель', separator, 'по которому тексты из всех ячеек сначала объединятся в один, а потом снова сепарируются' 
              , '\n--- Придумайте другой разделитель, не забывая об особенностях спецсимволов, впишите его и нажмите Enter')        
        separator = input()
    textS = df[columnWithText].tolist()
    textS = mstem.lemmatize(separator.join(textS))
    textS = ' '.join(textS)
    df[columnWithText] = textS.split(separator)
    df[columnWithText] = df[columnWithText].apply(lambda text: re.sub(r'  +', ' ', text))
    return df[columnWithText]

def simbolsCleaner(text):
    """
    Функция для чистки текстов от невербального мусора (ненужных символов)
    
    Parameters
    ----------
    text : str
    """
    time.sleep(0.01)
    print('                                                                                          ', end='\r')
    cleaned_text = ''
    for a in text:
        if (a.isalnum()) | (a == ' '): cleaned_text += a
            
    while '  ' in cleaned_text: cleaned_text = cleaned_text.replace('  ', ' ')
    while cleaned_text[0] == ' ': cleaned_text = cleaned_text[1:] # избавиться от пробелов в начале текста 
    while cleaned_text[-1] == ' ': cleaned_text = cleaned_text[:-1] # избавиться от пробелов в конце текста 
    return cleaned_text

def stopwordsDropper(text):
    """
    Функция для чистки текстов от стоп-слов пакетом stop_words
    
    Parameters
    ----------
    text : str
    """
    time.sleep(0.01)
    print('                                                                                          ', end='\r')
    stopwords_list = stop_words.get_stop_words('russian')
    text_cleaned = ''
    for word in text.split(' '):
        if word not in stopwords_list:
            text_cleaned += ' ' + word
    text_cleaned = text_cleaned.strip()
    # text_cleaned = text_cleaned.split(' ')
    return text_cleaned



#!/usr/bin/env python
# coding: utf-8

'''
A module for preprocessing variables of nominal, ordinal, interval, and higher-level measurement
'''

# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        import matplotlib.pyplot as plt, pandas
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '') #.replace('_', '')
        if '.' in module: module = module.split('.')[1] 
        print('Пакет', module,
              'НЕ прединсталлируется с установкой Анаконды, но для работы скрипта требуется этот пакет, поэтому он будет инсталлирован сейчас\n')
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print('Пакет', module
                  , 'НЕ прединсталлируется с установкой Анаконды, для работы скрипта требуется этот пакет,'
                  , 'но инсталлировать его не удаётся, попробуйте инсталлировать его вручную, после чего снова запустите требуемый скрипт пакета\n')
            break

def varHist(dfIn, var):
    plt.figure()
    plt.hist(dfIn[var])
    plt.title(f'Values distribution of {var}')
    plt.xlabel(f'{var}')
    plt.xticks(dfIn[var].unique(), minor=True)
    plt.ylabel('Frequency')
    plt.show();    

def valuesDropper(dfIn, var):
    df = dfIn.copy()
    print('В этой функции переменная рассматривается как порядковая или более высокого типа шкалы')
    varHist(df, var)
    print('\n--- Если требуется исключить из дальнейшего рассмотрения какие-то значения переменной, введите в окно МИНИМАЛЬНОЕ из этих значений и нажмите Enter'
        , '\n--- В противном случае, просто нажмите Enter')
    valueForRemoving = input()
    if len(valueForRemoving) != 0:
        valueS_ForRemoving = [valueForRemoving]
        print('\n--- Теперь введите МАКСИМАЛЬНОЕ из этих значений и нажмите Enter'
            , '\n--- Если максимум равен минимуму, просто нажмите Enter')
        valueForRemoving = input()
        if len(valueForRemoving) != 0:
            valueS_ForRemoving.append(valueForRemoving)
            print('\n--- Если ВСЕ числа в обзначенном Вами диапазоне требуется исключить из дальнейшего рассмотрения, просто нажмите Enter'
                , '\n--- В противном случае введите конкретное число из этого диапазона и нажмите Enter')
            valueForRemoving = input()
            if len(valueForRemoving) != 0:
                while len(valueForRemoving) != 0:
                    valueS_ForRemoving.append(valueForRemoving)
                    print('\n--- Если требуется исключить из дальнейшего рассмотрения ЕЩЁ одно конкретное число из этого диапазона, введите его и нажмите Enter'
                        , '\n--- В противном случае просто нажмите Enter')
                    valueForRemoving = input()
                    valueS_ForRemoving.append(valueForRemoving)
                # Исключить из рассмотрения диапазон
                valueS_ForRemoving.remove('')
                for valueForRemoving in valueS_ForRemoving:
                    df = df[dataForML[var] != float(valueForRemoving)]
            else:
                # Исключить из рассмотрения диапазон
                valueS_ForRemoving.sort()
                df = df[(df[var] < float(valueS_ForRemoving[0])) | (df[var] > float(valueS_ForRemoving[-1]))]
    
        print('\nДля дальнейшего рассмотрения остались следующие значения игрека и их частоты:')
        varHist(df, var)
    return df

# # 1.1.3 Если требуется выбрать только интересующие значения игрека
# print('В этом чанке игрек рассматривается как номинальная переменная'
#     , '\n--- Если требуется объединить какие-то значения игрека в одно, впишите в окно эти значения без кавычек через запятую с пробелом и нажмите Enter'
#     , '\n--- В противном случае, просто нажмите Enter\n')
# dataForML[varY] = dataForML[varY].astype(str)
# valueS_ForJoining = input()
# if len(valueS_ForJoining) != 0:
#     while len(valueS_ForJoining) != 0:
#         valueS_ForJoining = valueS_ForJoining.split(', ')
#         labelForJoined = input('--- Введите в окно метку для обозначения этих значений и нажмите Enter')
#         for valueForJoining in valueS_ForJoining:
#             print('Поглащается значение', valueForJoining)
#             dataForML.loc[dataForML[varY] == valueForJoining, varY] = labelForJoined
#         print('--- Если требуется объединить ещё какие-то значения игрека в одно, введите в окно эти значения через запятую и пробел, после чего нажмите Enter'
#             , '\n--- В противном случае, просто нажмите Enter\n')
#         valueS_ForJoining = input()

#     print('\nДля дальнейшего рассмотрения остались следующие значения игрека и их частоты:')
#     display(dataForML[varY].value_counts())
#     display(dataForML[varY].value_counts(normalize=True))
#     dataForML[varY].value_counts().plot.bar()



#TODO: handle binary str variables (as in logistic regression)

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from itertools import combinations
from IPython.display import display

from sklearn.cluster import KMeans as KM_Base
from sklearn.metrics import silhouette_samples, silhouette_score

from .comparison_of_central_tendency import ANOVA
from scipy.spatial.distance import pdist, squareform


class KMeans:
    """
    Class to perform k-means cluster analysis.
    
    Parameters
    ----------
    n_clusters : int 
        Number of clusters in a solution

    Attributes
    ----------
    n_iters : int
        Number of iterations the algorithm used
    is_converged : bool
        Whether the algorithm was converged
    cluster_centers : pd.DataFrame
        Centers of obtained clusters
    inertia : float
        Sum of squared distances of observations to their cluster center
    """
    def __init__(
        self,
        n_clusters=None,
    ):
        if n_clusters is None:
            raise ValueError('Please specify a number of clusters.')
        
        self.n_clusters = n_clusters
        
    def fit(
        self,
        data,
        variables=None,
        max_iter=100,
        tolerance=0.0001,
        random_state=1,
        show_results=True,
        n_decimals=3,
        plot_clusters=True,
        plot_silhouette=False
    ):
        """
        Fit a model to the given data.
        
        Parameters
        ----------
        data : pd.DataFrame 
            Data to fit a model
        variables : list 
            Names of variables from data that should be used in a cluster model.
            If not specified, all variables from data are used. Variables should have a numeric dtype.
        max_iter : int 
            Maximum number of iterations after which the algorithm is forced to stop
        tolerance : float 
            Minimum change in clusters' centers after which the algorithm is forced to stop
        random_state : int 
            An initial point of a random numbers generator to assure results' reproducibility
        show_results : bool 
            Whether to show results of analysis
        n_decimals : int 
            Number of digits to round results when showing them 
        plot_clusters : bool 
            Whether to visualize clusters in a dimension of two most influential variables 
        plot_silhouette : bool 
            Whether to visualize silhouette coefficient (looks like in SPSS' Two-Step Clusters)

        Returns
        -------
        self
            The current instance of the KMeans class
        """
        self._data = data.copy()
        
        if variables is None:
            data = data.dropna().copy()
            variables = list(data.columns)
        else:
            if not isinstance(variables, list):
                raise TypeError(f'Variables should be passed as list. {type(variables)} was passed instead.')
            
            data = data[variables].dropna().copy()
        
        self._integer_type_vars = [var for var in variables if (data[var].apply(round)!=data[var]).sum()==0]
            
        self._observation_idx = list(data.index)
        self._variables = list(data.columns)
        self._cluster_idx = [i+1 for i in range(self.n_clusters)]
        
        self._model = KM_Base(
            n_clusters=self.n_clusters,
            init='k-means++',
            n_init=10, 
            max_iter=max_iter,
            tol=tolerance,
            algorithm='full',
            random_state=random_state
        ).fit(data)
        
        self.n_iters = self._model.n_iter_
        self.is_converged = True if self.n_iters < max_iter else False
        self.cluster_centers = self.get_cluster_centers(round_discrete=False)
        self.inertia = self._model.inertia_
        self._predictions = self.transform(self._data, add_to_data=True)        
        
        if show_results:
            self.show_results(n_decimals)
        
        if plot_silhouette:
            print('------------------\n')
            print('Clusters quality')
            self.get_silhouette_plot()
            print("""This plot is based on the average silhouette score, 
which can be obtained by using [model].get_silhouette_score()""")

        if plot_clusters:
            print('------------------\n')
            print('Clusters visualisation')
            two_most_infl_vars = self._identify_two_influential_vars()
            self.get_bivariate_plots(two_most_infl_vars)
            print("""Only two most influential variables are shown (based on F-statistic's p-value).
To see more plots, use [model].get_bivariate_plots()""")
        
        return self
    
    def show_results(self, n_decimals=3):
        """
        Show results of the analysis in a readable form.
        
        Parameters
        ----------
        n_decimals : int 
            Number of digits to round results when showing them
        """
        phrase = 'method {}'
        
        print('\nK-MEANS SUMMARY')
        print('------------------\n')
        if self.is_converged:
            print(f'Estimation was succesfully converged in {self.n_iters} iterations.')
        else:
            print('Estimation was NOT converged successfully.')
            print('Please enlarge the number of iterations.')
        print('------------------\n')
        print('Final cluster centers')
        display(self.get_cluster_centers(round_discrete=True).style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption(phrase.format('.get_cluster_centers()')))
        if self._integer_type_vars != []:
            print("""Cluster centers for discrete variables are rounded for better interpretability.
To see the exact centers, use [model].get_cluster_centers(round_discrete=False)""")
        print('------------------\n')
        print('Distances between centers')
        display(self.get_distances_between_centers().style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption(phrase.format('.get_distances_between_centers()')))
        print('------------------\n')
        print('ANOVA')
        display(self.get_ANOVA_table().style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption(phrase.format('.get_ANOVA_table()')))
        print('------------------\n')
        print('Cluster membership')
        display(self.get_number_of_cases_by_clusters().style\
                    .format(None, na_rep="", precision=n_decimals)\
                    .set_caption(phrase.format('.get_number_of_cases_by_clusters()')))

    
    def get_cluster_centers(self, round_discrete=False):
        """
        Get the dataframe with cluster centers (means of each variable within clusters).
        
        Parameters
        ----------
        round_discrete : bool 
            Whether to round values of discrete variables for better interpretability

        Returns
        -------
        pd.DataFrame
            A table with cluster centers
        """
        
        centers = self._model.cluster_centers_
        variables = self._variables
        idx = self._cluster_idx
        
        result = pd.DataFrame(
            centers,
            columns=variables,
            index = idx
        )
        
        if round_discrete:
            for var in self._integer_type_vars:
                result[var] = result[var].apply(round)
        
        result.index.name = 'Cluster'
        
        return result
    
    def get_number_of_cases_by_clusters(self):
        """
        Get the frequency table, illustrating the distribution of observations by clusters.
        
        Returns
        -------
        pd.DataFrame
            A table with number of observations in each cluster
        """
        preds = self._predictions.copy()
        dist = pd.DataFrame(
            preds['Cluster'].value_counts(dropna=False).sort_index()
        )
        dist.columns = ['N']
        if len(dist.index) > len(self._cluster_idx):
            dist.index = self._cluster_idx + ['Missing']
        else:
            dist.index = self._cluster_idx
        dist.index.name = 'Cluster'
        dist.loc['Valid'] = dist.loc[self._cluster_idx, 'N'].sum()
        
        return dist.T
    
    def get_ANOVA_table(self):
        """
        Get the results of ANOVA analysis, i.e. analysis of differences in means of each variable between clusters.

        Returns
        -------
        pd.DataFrame
            A table with results of ANOVA test
        """
        preds = self._predictions.copy()
        summary = ANOVA(preds, self._variables, 'Cluster', show_results=False).summary()
        idx_to_show_b = [i for i in summary.index if 'Between Groups' in i]
        idx_to_show_w = [i for i in summary.index if 'Within Groups' in i]
        cols_to_show = ['Mean Square', 'df', 'F', 'p-value']
        summary_b = summary.reindex(idx_to_show_b)[cols_to_show]
        summary_w = summary.reindex(idx_to_show_w)[cols_to_show]
        summary_b.index = self._variables
        summary_w.index = self._variables
        summary = pd.concat([summary_b, summary_w], axis=1)
        summary.columns = [
            'Mean Square (Between Groups)',
            'df (Between Groups)',
            'F',
            'p-value',
            'Mean Square (Within Groups)',
            'df (Within Groups)',
            'F not used',
            'p-value not used']
        columns_to_show = [
            'Mean Square (Between Groups)',
            'df (Between Groups)',
            'Mean Square (Within Groups)',
            'df (Within Groups)',
            'F',
            'p-value']
        return summary[columns_to_show]
    
    def get_distances_between_centers(self):
        """
        Get Eucledian's distance between each cluster's centers.

        Returns
        -------
        pd.DataFrame
            A table with distances between each cluster
        """
        centers = self.cluster_centers
        idx = list(centers.index)
        dists = squareform(pdist(centers))
        dists = pd.DataFrame(dists, index=idx, columns=idx)
        dists.index.name = 'Cluster'
        return dists
        
    def transform(
        self,
        data=None,
        cluster_membership=True,
        distance_to_centers=False,
        add_to_data=False
    ):
        """
        Assign each observation to a relevant cluster and / or estimate distance between each observation 
        and all clusters' centers.
        
        Parameters
        ----------
        data : pd.DataFrame 
            Data for which new variables are estimated.
            If not specified, data that were used to fit a model are used.
        cluster_membership : bool 
            Whether to return cluster membership for each observation
        distance_to_centers : bool 
            Whether to return distance to cluster centers for each observation
        add_to_data : bool 
            Whether to merge new values with the given data

        Returns
        -------
        pd.DataFrame
            Requested transformations
        """        
        if data is None:
            data_copy = self._data.copy()
            transform_data = self._data[self._variables].dropna().copy()
        else:
            data_copy = data.copy()
            transform_data = data[self._variables].dropna().copy()
        
        observation_idx = list(transform_data.index)
        
        cl_memb = self._model.predict(transform_data) + 1
        dist = self._model.transform(transform_data)    
        
        cl_memb = pd.DataFrame(
            cl_memb,
            index=observation_idx,
            columns=['Cluster'],
            dtype=int
        )
        
        dist_cols = [f'Distance to cluster {i}' for i in self._cluster_idx]
        dist = pd.DataFrame(
            dist,
            index=observation_idx,
            columns=dist_cols
        )
        
        columns_to_show = []
        
        if add_to_data:
            columns_to_show.extend(list(data_copy.columns))
        if cluster_membership:
            columns_to_show.append('Cluster')
        if distance_to_centers:
            columns_to_show.extend(dist_cols)      
        
        data_copy['Cluster'] = cl_memb
        data_copy[dist_cols] = dist        

        result = data_copy[columns_to_show].copy()
        
        return result
    
    def get_bivariate_plots(self, variables=None):
        """
        Visualize clusters in a bivariate dimension of all or the requested variables.
        
        Parameters
        ----------
        variables : list
            List of names of variables for which plots should be shown.
            If not specified, all variables that were used to fit a model are used.
        """ 
        if variables is None:
            variables = self._variables
        combos = list(combinations(variables, 2))
        data = self._predictions[variables + ['Cluster']]
        centers = self.cluster_centers[variables]
        for combo in combos:
            KMeans._one_bivariate_cluster_plot(data, combo, centers[list(combo)])
    
    @staticmethod
    def _one_bivariate_cluster_plot(data, two_variables, centers):
        var1, var2 = two_variables
        data = data[list(two_variables) + ['Cluster']].dropna()
        plt.figure(figsize=(8, 5))
        plt.scatter(data[var1], data[var2], c=data['Cluster'], cmap='Set3')
        plt.scatter(centers.loc[:, var1], centers.loc[:, var2], marker='o',
                    c="white", alpha=1, s=200, edgecolor='k')
        for i, c in enumerate(centers.values):
            k = i + 1
            plt.scatter(c[0], c[1], marker='$%d$' % k, alpha=1,
                        s=50, edgecolor='k')
        plt.xlabel(var1)
        plt.ylabel(var2)
        plt.title('Cluster membership')
        plt.show();
        
    def _identify_two_influential_vars(self):
        anv = self.get_ANOVA_table()
        anv = anv.sort_values('p-value')
        idx_infl = list(anv.index)[:2]
        return idx_infl
    
    def get_silhouette_score(self):
        """
        Return a value of the average silhouette score for the current model.
        
        Returns
        -------
        float
            A silhouette coefficient
        """
        data = self._predictions[self._variables].dropna()
        labels = np.ravel(self._predictions['Cluster'].dropna())
        return silhouette_score(data, labels)
    
    def get_silhouette_plot(self):
        """
        Visualize the model's quality based on its average silhouette score.
        
        """
        silh_score = self.get_silhouette_score()
        silh_point = 1 + silh_score
        plt.figure(figsize=(8, 2))
        plt.barh(0.1, 2, left=-1, height=0.2, color='lightgreen', alpha=0.5)
        plt.barh(0.1, 1.5, left=-1, height=0.2, color='papayawhip', alpha=0.6)
        plt.barh(0.1, 1.2, left=-1, height=0.2, color='salmon', alpha=0.7)
        plt.barh(0.1, silh_point, left=-1, height=0.1, color='orchid', edgecolor='black')

        plt.scatter([silh_score], [0.1], c='black', zorder=2)
        plt.annotate(round(silh_score, 2), (silh_score+0.01, 0.1+0.01))
        plt.annotate('Poor', (0.07, 0.01))
        plt.annotate('Fair', (0.37, 0.01))
        plt.annotate('Good', (0.85, 0.01))
        plt.yticks([])
        plt.title('Solution quality')
        plt.show();



#!/usr/bin/env python
# coding: utf-8

'''
A module for working on a calendar within a specific year
'''
# import sys
# sys.path.append(r"C:\Users\Alexey\Dropbox\Мои\RAnDan\myModules")

# sys & subprocess -- эти пакеты должны быть предустановлены. Если с ними какая-то проблема, то из этого скрипта решить их сложно
import sys
from subprocess import check_call

# --- остальные модули и пакеты
attempt = 0
while True:
    try:
        from pandas import DataFrame
        break
    except ModuleNotFoundError:
        errorDescription = sys.exc_info()
        module = str(errorDescription[1]).replace("No module named '", '').replace("'", '') #.replace('_', '')
        print('Пакет', module, 'НЕ прединсталлируется с установкой Анаконды, но для работы скрипта требуется этот пакет, поэтому он будет инсталлирован сейчас\n')
        check_call([sys.executable, "-m", "pip", "install", module])
        attempt += 1
        if  attempt == 10:
            print('Пакет', module
                  , 'НЕ прединсталлируется с установкой Анаконды, для работы скрипта требуется этот пакет,'
                  , 'но инсталлировать его не удаётся, попробуйте инсталлировать его вручную, после чего снова запустите требуемый скрипт пакета\n')
            break

def calendarWithinYear(year):
    """
    Функция "Календарь день-месяц" для лучшей визуализации дальнейшего процесса
    
    Parameters
    ----------
    Аргументы этой функции аналогичны аргументам метода https://dev.vk.com/ru/method/newsfeed.search
    year : int
    """
    calendar = DataFrame()
    mnth_n31 = [2, 4, 6, 9, 11] # месяцы с числом дней менее 31
    for mnth in range(1, 13):
        if mnth == 2: # обработка февраля
            for day in range(1, 29):
                calendar.loc[str(day) if len(str(day)) == 2 else '0' + str(day), '02'] = 1
            if  year % 4 == 0: # обработка високосных лет
                calendar.loc['29', '02'] = 1
        elif (mnth in mnth_n31) & (mnth != 2): # месяцы с числом дней менее 31, но не февраль
            for day in range(1, 31):
                calendar.loc[str(day) if len(str(day)) == 2 else '0' + str(day)
                             , str(mnth) if len(str(mnth)) == 2 else '0' + str(mnth)] = 1
        else: # месяцы с числом дней 31
            for day in range(1, 32):
                calendar.loc[str(day) if len(str(day)) == 2 else '0' + str(day)
                             , str(mnth) if len(str(mnth)) == 2 else '0' + str(mnth)] = 1
    return calendar

def yearsRangeParser(yearsRange):
    yearsRange.sort()
    yearMinByUser = int(yearsRange[0])
    yearMaxByUser = int(yearsRange[-1])
    yearsRange = f'{yearMinByUser}-{yearMaxByUser}'
    return yearMaxByUser, yearMinByUser, yearsRange
